//
//  M3XViewController.m
//  microMemex
//
//  Created by narita on 2017/05/10.
//  Copyright © 2017年 narita. All rights reserved.
//

#import "M3XViewController.h"
#import "M3XNote+CoreDataClass.h"
#import "M3XNote+CoreDataProperties.h"

@implementation M3XViewController

#pragma mark - class methods

#pragma mark - init methods

#pragma mark - dealloc

#pragma mark - NSCopying, hash, isEqual:

#pragma mark - NSCoding methods

#pragma mark - restorableState methods

+ (NSArray *)restorableStateKeyPaths
{
    return [[super restorableStateKeyPaths] arrayByAddingObjectsFromArray:@[@"selectedIndexPath"]];
}

#pragma mark - life cycle methods

- (void)viewDidLoad
{
    [super viewDidLoad];
    
    // Do any additional setup after loading the view.  
    
    // ここでは、まだ"self.representedObject"の値は設定されていない。
    
    // Dropを受け入れるデータの型の登録
    [self.browserView registerForDraggedTypes:@[[M3XNote pasteBoardRowsType]]];
    
    [self.browserView setDraggingSourceOperationMask:NSDragOperationLink
                                            forLocal:YES];

    
    // restorableStateのための小細工
    [[NSOperationQueue mainQueue] addOperationWithBlock:^
     {
         if( self.selectedIndexPath == nil )
         {
             self.selectedIndexPath = [NSIndexPath indexPathWithIndex:0];
         }
         
         [self trySetBrowserViewSelectionIndexPath:self.selectedIndexPath];
     }];    
}

- (void)viewWillAppear
{
    [super viewWillAppear];
    
    [self.browserView addObserver:self
                       forKeyPath:@"selectionIndexPaths"
                          options:(NSKeyValueObservingOptionNew | NSKeyValueObservingOptionOld )
                          context:(void*)self];
}


- (void) viewWillDisappear
{
    [self.browserView removeObserver:self forKeyPath:@"selectionIndexPaths"];
    
    [super viewWillDisappear];
}


#pragma mark - action methods

- (IBAction) new:(id)sender
{
    NSIndexPath* theSelectedPath = self.browserView.selectionIndexPath;
    
    // 選択されているitemの取得
    M3XNote* theSelectedNote = [[self rootItemForBrowser:self.browserView] noteFromIndexPath:theSelectedPath];
    
    M3XNote* theNewNote= [NSEntityDescription insertNewObjectForEntityForName:@"M3XNote" 
                                            inManagedObjectContext:self.managedObjectContext];
    
    NSDateFormatter* theDateFormatter = [[NSDateFormatter alloc] init];
    [theDateFormatter setDateFormat:@"yy/MM/dd H:mm:ss"];
    
    NSString* theMemoString = [NSString stringWithFormat:@"メモ : %@", [theDateFormatter stringFromDate:[NSDate date]]];

    NSAttributedString* theMemo = [[NSAttributedString alloc] initWithString:theMemoString];
    
    theNewNote.memo = [theMemo RTFDFromRange:NSMakeRange(0, theMemo.length)
                                    documentAttributes:@{}];
    theNewNote.isHome = NO;

    [theSelectedNote addLinksObject:theNewNote];
    
    // 選択箇所を一歩前進
    [[NSOperationQueue mainQueue] addOperationWithBlock:^
     {
         [self trySetBrowserViewSelectionIndexPath:
          [theSelectedPath indexPathByAddingIndex:[theSelectedNote.sortedLinks indexOfObject:theNewNote]]];
     }];    
}

- (BOOL) canNew:(nullable id<NSValidatedUserInterfaceItem>)item
{
    return self.browserView.selectionIndexPath != nil;;
}

- (IBAction) cut:(id)sender
{
    [self copy:sender];
    [self delete:sender];
}

- (BOOL) canCut:(nullable id<NSValidatedUserInterfaceItem>)item
{
    return self.browserView.selectionIndexPath != nil;;
}

- (IBAction) copy:(id)sender
{
    NSIndexPath* theSelectedPath = self.browserView.selectionIndexPath;

    // 選択されているitemの取得
    M3XNote* theSelectedNote       = [[self rootItemForBrowser:self.browserView] 
                                      noteFromIndexPath:theSelectedPath];
    
    // ペーストボードへ書き込み
    [theSelectedNote wreiteToPasteboard:[NSPasteboard generalPasteboard]];
}

- (BOOL) canCopy:(nullable id<NSValidatedUserInterfaceItem>)item
{
    return self.browserView.selectionIndexPath != nil;;
}

- (IBAction) paste:(id)sender
{
    NSIndexPath* theSelectedPath = self.browserView.selectionIndexPath;
    
    // 選択されているitemの取得
    M3XNote* theSelectedNote = [[self rootItemForBrowser:self.browserView] 
                                noteFromIndexPath:theSelectedPath];
    
    // ここでCoreDataに追加する
    M3XNote* theNoteFormPasteboard = [M3XNote readFromPasteboard:[NSPasteboard generalPasteboard]
                                          inManagedObjectContext:self.managedObjectContext];
    
    [theSelectedNote addLinksObject:theNoteFormPasteboard];
}

- (BOOL) canPaste:(nullable id<NSValidatedUserInterfaceItem>)item
{
    NSIndexPath* theSelectedPath = self.browserView.selectionIndexPath;
    
    // 選択されているitemの取得
    M3XNote* theSelectedNote       = [[self rootItemForBrowser:self.browserView] 
                                      noteFromIndexPath:theSelectedPath];

    // もう少しスマートな書き方があるはず
    NSPasteboard*   thePasteboard = [NSPasteboard generalPasteboard];
    NSData*         theData = [thePasteboard dataForType:[M3XNote pasteBoardRowsType]];

    return (theSelectedNote != nil) && (theData != nil);
}


- (IBAction) delete:(id)sender
{
    NSIndexPath* theSelectedPath = self.browserView.selectionIndexPath;
    
    if( theSelectedPath != nil )
    {
        // 選択されているitemの取得
        M3XNote* theSelectedNote       = [[self rootItemForBrowser:self.browserView] 
                                          noteFromIndexPath:theSelectedPath];
        
        // 選択されているitemの左側のitemを取得。仮に"親"とする。
        M3XNote* theSelectedParentNote = [[self rootItemForBrowser:self.browserView] 
                                          noteFromIndexPath:[theSelectedPath indexPathByRemovingLastIndex]];
        
        // 選択箇所を一歩後退
        self.browserView.selectionIndexPaths = @[ [theSelectedPath indexPathByRemovingLastIndex] ];

        // 親のリンクからを削除
        [theSelectedParentNote removeLinksObject:theSelectedNote];
        
        // どこからもリンクされていない場合は、本体を削除
        // この場合は、ルートへの参照は無視しない。
        if( theSelectedNote.links.count == 0 ) 
        {
            [self.managedObjectContext deleteObject:theSelectedNote];
        }                
    }
}

- (BOOL) canDelete:(nullable id<NSValidatedUserInterfaceItem>)item
{
    return self.browserView.selectionIndexPath != nil;;
}

// ユーザークリックによる選択箇所の変化の取得。(何故かselectionIndexPathsのKVOが効かないので、ここで強制させる)
- (IBAction)browserSelectionDidChange:(id)sender
{
    self.browserView.selectionIndexPaths = self.browserView.selectionIndexPaths;
}

// ツールバーやメニューバーには効くがNSButtonには有効ではない。
- (BOOL)validateUserInterfaceItem:(id<NSValidatedUserInterfaceItem>)item
{
    BOOL theResult = NO;
    SEL  theAction = [item action];
    
    if( [self respondsToSelector:theAction] )
    {
        NSString* theActionString = NSStringFromSelector(item.action);
        
        // "actionName:"を"canActionName:"にする。
        NSString* theCanActionString = [NSString stringWithFormat:@"can%@%@",
                                        [[theActionString substringToIndex:1] uppercaseString],// 頭文字は大文字に
                                        [theActionString substringFromIndex:1]]; // 頭文字以外はそのまま使う
        
        SEL theCanActon = NSSelectorFromString(theCanActionString);
        
        if( [self respondsToSelector:theCanActon] )
        {
            // やってることは[performSelector:withObjecy:]とほぼ同じ。返り値がBOOLのためにポインターを抜き出している
            typedef BOOL  (*methodFuncPtr)(id, SEL, id);
            methodFuncPtr theFunction = nil;
            
            theFunction = (methodFuncPtr)[ self methodForSelector:theCanActon];
            theResult   = (*theFunction)(self, theCanActon, item);
        }
        else
        {
            theResult = YES;
        }
    }
    else
    {
        // superにはvalidateUserInterfaceItem:がないので無視
    }    
    
    return theResult;
}

#pragma mark - event handling methods

- (void) updateBrowserViewAndTextView:(NSNotification*)inNotification
{
    // selectionを退避
    NSIndexPath* theSelectedPath = self.browserView.selectionIndexPath;

    self.browserView.selectionIndexPaths = @[];
    [self.browserView validateVisibleColumns];    
    
    if( theSelectedPath != nil )
    {
        [self trySetBrowserViewSelectionIndexPath:theSelectedPath];
    }
}

- (void)observeValueForKeyPath:(nullable NSString *)keyPath
                      ofObject:(nullable id)object
                        change:(nullable NSDictionary<NSKeyValueChangeKey, id> *)change
                       context:(nullable void *)context
{
    if( object == self.browserView )
    {
        if( [keyPath isEqualToString:@"selectionIndexPaths"] )
        {
            // browserViewの選択箇所が変わるたびに、TextViewの表示を変えるため。
            // このselectionNodeをCocoa bindingでNSTextViewへ接続する
            self.selectedNode = [[self rootItemForBrowser:
                                  self.browserView] noteFromIndexPath:self.browserView.selectionIndexPath];
            
            // restorableStateのための小細工
            self.selectedIndexPath = self.browserView.selectionIndexPath;
            [self invalidateRestorableState];            
        }
    }
}

#pragma mark - drawing methods

#pragma mark - delegate/datasource methods

#pragma mark - NSTextViewDelegate methods

- (BOOL)textView:(NSTextView *)textView doCommandBySelector:(SEL)commandSelector
{
    if( commandSelector == @selector(insertTab:) )
    {
        [self.view.window makeFirstResponder:self.browserView];
        
        return YES;
    }
         
    return NO;
}

#pragma mark - NSBrowserDelegate methods

// Drag&DrilDown時にDrop先のTextViewへ内容へ変更する小細工
- (void)browser:(NSBrowser *)browser didChangeLastColumn:(NSInteger)oldLastColumn toColumn:(NSInteger)column
{    
    self.selectedNode = [[self rootItemForBrowser:
                          self.browserView] noteFromIndexPath:self.browserView.selectionIndexPath];
}

- (id)rootItemForBrowser:(NSBrowser *)browser
{    
    NSError* theError = nil;
    
    NSArray* theResult = [self.managedObjectContext executeFetchRequest:[M3XNote rootFetchRequest]
                                                                  error:&theError];
    
    return theResult.firstObject;
}

- (NSInteger)browser:(NSBrowser *)browser numberOfChildrenOfItem:(id)item
{
    M3XNote* theNote = (M3XNote *)item;
    
    return theNote.sortedLinks.count;
}

- (id)browser:(NSBrowser *)browser child:(NSInteger)index ofItem:(id)item
{
    M3XNote* theNote = (M3XNote *)item;
        
    return theNote.sortedLinks[index];        
}

- (BOOL)browser:(NSBrowser *)browser isLeafItem:(id)item
{
    M3XNote* theNote = (M3XNote *)item;
    
    return theNote.sortedLinks.count == 0;
}

- (id)browser:(NSBrowser *)browser objectValueForItem:(id)item
{
    M3XNote* theNote = (M3XNote *)item;
        
    return theNote.titleString;
}

// Drag元
// 以下でYESを返せばDrag元になれる
- (BOOL)browser:(NSBrowser *)browser writeRowsWithIndexes:(NSIndexSet *)rowIndexes inColumn:(NSInteger)column toPasteboard:(NSPasteboard *)pasteboard
{
    // pasteboardにデータをNoteのObjectID入れること
    M3XNote* theParentNote =  [self.browserView itemAtIndexPath:
                               [self.browserView indexPathForColumn:column]];

    // Dragできるのは一つだけ
    M3XNote* theDraggedNote = theParentNote.sortedLinks[rowIndexes.firstIndex];
    
    // 書き込み
    [theDraggedNote wreiteToPasteboard:pasteboard];
    
    return YES;
}

// Drop先
- (NSDragOperation)browser:(NSBrowser *)browser
              validateDrop:(id <NSDraggingInfo>)info
               proposedRow:(NSInteger *)row
                    column:(NSInteger *)column
             dropOperation:(NSBrowserDropOperation *)dropOperation
{
    // item間やitem以外の場所へのDropはカラム全体へのDropと見なす
    if( *dropOperation == NSBrowserDropAbove )
    {
        *row = -1;
        *dropOperation = NSBrowserDropOn;
    }
    
    if( *dropOperation == NSBrowserDropOn )
    {
        // NSDraggingInfoから
        if( info.draggingSource == self.browserView )
        {
            return NSDragOperationLink;            
        }
        else
        {
            return NSDragOperationCopy;            
        }
    }
    
    return NSDragOperationNone;            
}

// Drop後の処理
- (BOOL)  browser:(NSBrowser *)browser
       acceptDrop:(id <NSDraggingInfo>)info
            atRow:(NSInteger)row
           column:(NSInteger)column
    dropOperation:(NSBrowserDropOperation)dropOperation
{    
    // まず、Drop先のNoteを求める。
    // pasteboardにデータをNoteのObjectID入れること
    M3XNote* theParentNote =  [self.browserView itemAtIndexPath:
                               [self.browserView indexPathForColumn:column]];
    
    M3XNote* theDropNote;
    
    if( row == -1 )// カラム全体
    {
        theDropNote = theParentNote;
    }
    else
    {
        // Dragできるのは一つだけ
        theDropNote = theParentNote.sortedLinks[row];
    }
    
    // ここでCoreDataに追加する
    M3XNote* theDraggedNote = [M3XNote readFromPasteboard:info.draggingPasteboard
                                   inManagedObjectContext:self.managedObjectContext];
    
    [theDropNote addLinksObject:theDraggedNote];

    return YES;        
}


#pragma mark - accessor methods (in pairs)

+ (NSSet *)keyPathsForValuesAffectingRepresentedObject
{
    return [NSSet setWithObjects:@"managedObjectContext", nil];
}

- (void)setRepresentedObject:(id)representedObject
{
    // cocoa bindingを諦めた代わりにchangeNotificationを取得してUIを更新する
    NSNotificationCenter* theNotificationCenter = NSNotificationCenter.defaultCenter;
    
    if( self.representedObject )
    {
        [theNotificationCenter removeObserver:self
                                         name:NSManagedObjectContextObjectsDidChangeNotification 
                                       object:self.representedObject];
    }
    
    [super setRepresentedObject:representedObject];    
    
    if( representedObject )
    {
        [theNotificationCenter addObserver:self
                                  selector:@selector(updateBrowserViewAndTextView:) 
                                      name:NSManagedObjectContextObjectsDidChangeNotification 
                                    object:representedObject];
    }
}

- (NSManagedObjectContext*) managedObjectContext
{
    return (NSManagedObjectContext*) self.representedObject;
}

#pragma mark - Utility methods

- (void) trySetBrowserViewSelectionIndexPath:(NSIndexPath*)indexPath
{
    // ルートから辿ってゆく可能なところまでselectionを設定
    // undo/redoで選択箇所が削除されている場合があるのでrootから選択出来るところでの選択にする
    M3XNote* theNote = [self rootItemForBrowser:self.browserView];
    NSUInteger theResultLength = 0;
    
    // 可能な長さを求めて
    for(NSUInteger thePosition = 0; thePosition < indexPath.length; thePosition++)
    {
        if( theNote.sortedLinks.count <= [indexPath indexAtPosition:thePosition] )
        {
            // 長さが足らないので今までの部分だけで選択させる
            theResultLength = thePosition;
            break;
        }
        
        theNote = theNote.sortedLinks[[indexPath indexAtPosition:thePosition]]; 
        theResultLength = thePosition + 1;        
    }
    
    // お尻から削って目的の長さにする。
    NSIndexPath* theResult = indexPath;
    
    while( theResult.length != theResultLength )
    {
        theResult = [theResult indexPathByRemovingLastIndex];
    }
    
    self.browserView.selectionIndexPaths = @[theResult];
}


@end
